"""
6502 code static analysis

Machine:
	RAM (64K buffer)
	Bus (address xlate)
	CPU
"""

class Error(Exception):
	pass

class BusError(Error):
	pass

def trace(msg):
	print msg

class RAM(object):
	def __init__(self, length):
		self.d = [0] * length
	
	def readb(self, addr):
		return self.d[addr]

	def writeb(self, addr, b):
		self.d[addr] = b

class Bus(object):
	def __init__(self, busConfiguration):
		self.addrmap = []
		for firstaddr, length, deviceClass in busConfiguration:
			device = deviceClass(length)
			self.addrmap.append( (firstaddr, length, device) )
	
	def readb(self, addr):
		firstAddr, device = self.getDevice(addr)
		return device.readb(addr - firstAddr)

	def writeb(self, addr, b):
		firstAddr, device = self.getDevice(addr)
		return device.writeb(addr - firstAddr, b)

	def getDevice(self, addr):
		for firstaddr, length, device in self.addrmap:
			if addr >= firstaddr and addr < (firstaddr + length):
				return firstaddr, device
		raise BusError("No device mapped at address %04x" % (addr))

AllRAMBusConfiguration = [(0, 64 * 1024 * 1024, RAM)]

class System(object):
	def __init__(self, busConfiguration = None):
		if busConfiguration is None:
			busConfiguration = AllRAMBusConfiguration

		self.bus = Bus(busConfiguration)
		self.readb = self.bus.readb
		self.writeb = self.bus.writeb

		self.exitOnBrk = False
		self.keepRunning = True

		self.reset()

	def reset(self):
		self.A = 0
		self.X = 0
		self.Y = 0
		self.S = 0
		
		# Flags
		self.NF = 0 # 1 = negative
		self.VF = 0 # 1 = overflow
		self.BF = 0 # 1 = brk
		self.DF = 0 # 1 = decimal
		self.IF = 0 # 1 = IRQ disable
		self.ZF = 0 # 1 = result zero
		self.CF = 0 # 1 = carry

		self.PC = self.readb(0xfffd) << 8
		self.PC |= self.readb(0xfffc)

	def setExitOnBrk(self, exitOnBrk = True):
		self.exitOnBrk = exitOnBrk
	
	def readpc(self):
		b = self.readb(self.PC)
		self.PC += 1
		return b

	def readpc2(self):
		b = self.readb(self.PC) | (self.readb(self.PC + 1) << 8)
		self.PC += 2
		return b

	def readimm(self, mnem):
		M = self.readpc()
		trace("%s #$%02x" % (mnem, M))
		return M

	def readzp(self, mnem):
		addr = self.readpc()
		trace("%s $%02x" % (addr))
		return self.readb(addr)

	def readzpindx(self, mnem):
		addr = self.readpc()
		trace("%s $%02,X" % (mnem, addr))
		return self.readb(addr + self.X)

	def readabs(self, mnem):
		addr = self.readpc2()
		trace("%s $%04x" % (mnem, addr))
		return self.readb(addr)

	def readabsx(self, mnem):
		addr = self.readpc2()
		trace('%s $%04,X' % (mnem, addr))
		if self.page_boundary_crossed(addr, self.X):
			self.cycles(1) # Additional cycle for page boundary
		return self.readb(addr + self.X)

	def readabsy(self, mnem):
		addr = self.readpc2()
		trace('%s $%04,Y' % (mnem, addr))
		if self.page_boundary_crossed(addr, self.Y):
			self.cycles(1) # Additional cycle for page boundary
		return self.readb(addr + self.Y)

	def readindx(self, mnem):
		idx = self.readpc()
		trace('%s ($%02, X)' % (mnem, idx))
		idx += self.X
		idx &= 0xff
		addr = readb(idx) | (readb(idx + 1) << 8)
		return self.readb(addr)

	def readindy(self, mnem):
		idx = self.readpc()
		trace('%s ($%02, Y)' % (mnem, idx))
		idx += self.X
		idx &= 0xff
		addr = readb(idx) | (readb(idx + 1) << 8)
		if idx + 1 == 256:
			self.cycles(1) # Page-boundary cross.
		return self.readb(addr)
	
	def run(self):
		self.keepRunning = True
		while self.keepRunning:
			opcode = self.readpc()
			handler = getattr(self, 'op_%02x' % (opcode))
			handler()
	
	def cycles(self):
		# Count instruction cycles
		pass

	def page_boundary_crossed(self, addr, idx):
		if (addr & 0xff) + idx > 256:
			return 1
		else:
			return 0

	# ADC
	def _adc(self, M):
		# Add with carry.
		assert self.DF == 0, "No decimal yet"

		self.A = self.CF + self.A + M
		self.C = (self.A > 0xff)
		self.A &= 0xFF
		self.ZF = (self.A == 0)


	def op_69(self):
		self.cycles(2)
		return self._adc(self.readimm('ADC'))
	
	def op_65(self):
		self.cycles(3)
		return self._adc(self.readzp('ADC'))
	
	def op_75(self):
		self.cycles(4)
		return self._adc(self.readzpindx('ADC'))

	def op_6d(self):
		self.cycles(4)
		return self._adc(self.readabs('ADC'))

	def op_7d(self):
		self.cycles(4)
		return self._adc(self.readabsx('ADC'))

	def op_79(self):
		self.cycles(4)
		return self._adc(self.readabsy('ADC'))

	def op_61(self):
		self.cycles(6)
		return self._adc(self.readindx('ADC'))

	def op_71(self):
		self.cycles(5)
		return self._adc(self.readindy('ADC'))

	def op_78(self):
		trace("SEI")
		self.IF = 1

	def op_a2(self):
		self.X = self.readpc()
		trace("LDX #$%02X" % (self.X))

