#!/usr/bin/env python

import sys
import struct
import time
import socket
import errno
import struct

from libsid import SIDFile, makeDriver
import sidconnection

#SERIAL = '/dev/tty.PL2303-0000201A'
#SERIAL = '/dev/tty.usbserial-A7005frV'
#SERIAL = '/dev/tty.SIDPLAYER-DevB'
SERIAL = '/dev/tty.GALWAY-DevB'

SERVERPORT=0x51DD

DEBUG_SID_INFO = True

# Instructions for programming the PIC timer1.
# Byte 0  q  Command "custom timer"
#      1  1  Enable timer
#      2  H  value of TMR1H
#      3  L  value of TMR1L
#      4  1  value of RD16
#      5  1  value of T1CKPS1
#      6  0  value of T1CKPS0
#      7  0  value of T1OSCEN (0 = INTOSC / 4)
CMD_50_HZ = 'q' + chr(1) + chr(0xb1) + chr(0xe0) + chr(1) + chr(1) + chr(0) + chr(0)
CMD_60_HZ = 'q' + chr(1) + chr(0xbe) + chr(0xe6) + chr(1) + chr(1) + chr(0) + chr(0)

class Error(Exception): pass
class UploadError(Error): pass
class ConnectionError(Error): pass

class ProxyDevice(object):
	def __init__(self):
		# Connect to the server
		self._con = None
		self.connectOrSpawn()

	def transact(self, dataIn):
		self._con.send(struct.pack('!H', len(dataIn)))
		self._con.send(dataIn)

		responseLen = self._con.recv(2)
		if not responseLen:
			raise ConnectionError("Server died?")
		responseLen = struct.unpack('!H', responseLen)[0]
		return self._con.recv(responseLen)

	def disconnect(self):
		if self._con:
			self._con.send(struct.pack('!H', 0))
			self._con.close()

	def spawnServer(self):
		print "Spawning server"
		server = sidconnection.SIDServer(SERIAL, SERVERPORT)
		server.spawn()
	
	def connectOrSpawn(self):
		def connect():
			s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			s.connect(('localhost', SERVERPORT))
			self._con = s

		triedSpawn = False
		while not triedSpawn:
			try:
				connect()
			except socket.error as e:
				if e.errno == errno.ECONNREFUSED:
					self.spawnServer()
					triedSpawn = True
			else:
				break

		if self._con == None:
			raise ConnectionError()

class PlaySID(object):
	""" Communicates with the SID player firmware (i.e., the program running
	on the PIC on SID player board) via a device (the serial port, right now
	-- see Device above)
	"""

	def __init__(self):
		self.sid = SIDFile()
		self.device = ProxyDevice()
	
	def done(self):
		self.device.disconnect()
		self.device = None

	def load(self, filename):
		self.sid.load(filename, printSIDInfo = DEBUG_SID_INFO)

	def _upload_page(self, num, data):
		assert len(data) <= 256
		cmd = 'u' + struct.pack('>HH', num, len(data)) + data
		# print "cmd", [hex(ord(c)) for c in cmd]
		response = self.device.transact(cmd)
		if not response.startswith('Uploaded'):
			print response
			raise UploadError()
		#print "Uploaded %02x: %r" % (num, response)
	
	def stop(self):
		cmd = 's'
		response = self.device.transact(cmd)
		#print "Stopped SID: %r" % (response)

	def upload(self):
		# Upload format: 256 bytes at a time, frame is:
		# uPPLLD
		# PP - 2 byte big-endian start address (lower 8 bits must be 0)
		# LL - 2 byte big-endian length
		# D  - LL bytes of data

		numPages = len(self.sid.pages)
		pageCount = 1
		for page in sorted(self.sid.pages.keys()):
			self._upload_page(page, self.sid.pages[page])
			percentage = (pageCount * 100) / numPages
			sys.stdout.write('\x1b[1K\r%d%%' % (percentage))
			sys.stdout.flush()
			pageCount += 1

		# Upload the hash also.
		cmd = 'n' + self.sid.songhash
		response = self.device.transact(cmd)

		print

		# Upload a hash of the song, so we know what's in memory.
	
	def upload_driver(self, songNumber):
		# Create the special ROM page which will:
		# a) serve as the reset vector pointer (0xfffc and 0xfffd)
		# b) serve as the IRQ vector pointer (the song play function)
		# c) contain startup code which:
		#    1) calls the song init function
		#    2) enables IRQs
		#    3) sits in a loop.
		self._upload_page(0xff00, makeDriver(self.sid.initAddr, self.sid.playAddr, songNumber))
	
	def hash_invalid(self):
		cmd = 'N'
		response = self.device.transact(cmd)
		return not response.startswith(self.sid.songhash)

	def play(self):
		if self.sid.refreshHz == 50:
			cmd = CMD_50_HZ
		elif self.sid.refreshHz == 60:
			cmd = CMD_60_HZ
		else:
			cmd = CMD_50_HZ # Default to PAL
		response = self.device.transact(cmd)

		print "Playing at %d Hz" % (self.sid.refreshHz)
		response = self.device.transact('p')
		
		#print "Played? %r" % (response)

def playSID(filename, songNumber, ignoreHash):
	player = PlaySID()
	player.stop()
	player.load(filename)
	if ignoreHash or player.hash_invalid():
		player.upload()
	player.upload_driver(songNumber)
	player.play()
	player.done()

def stop():
	player = PlaySID()
	player.stop()
	player.done()

def main():
	ignoreHash = False
	argv = sys.argv[:]
	if argv[1] == '--cold':
		ignoreHash = True
		argv[1:] = argv[2:]
	elif argv[1] == '--stop':
		stop()
		return
	elif argv[1] == '--kill':
		kill()
		return

	if len(argv) == 3:
		songNumber = int(argv[2])
	else:
		songNumber = 0
	playSID(argv[1], songNumber, ignoreHash)


if __name__ == '__main__':
	main()
